from copy import copy

import humps.main as humps
from ariadne import FallbackResolversSetter
from graphql.type import GraphQLList, GraphQLField
from .types import error_detail

__resolvers__ = ["field_error_resolver"]


@error_detail.field("error")
def resolve_error_detail_error(parent, info, *args, **kwargs):
    """Cast the error as a string to retrieve the message"""
    return str(parent)


def resolve_field_errors(parent, info, *args, **kwargs):
    fields = list()

    def traverse(node, stack=""):
        for name, errors in node.items():
            inner_stack = copy(stack)
            if bool(inner_stack):
                if isinstance(name, int):
                    inner_stack = f"{inner_stack}[{name}]"
                else:
                    inner_stack = f"{inner_stack}.{name}"
            else:
                inner_stack = name

            if isinstance(errors, list):
                # final condition
                fields.append(dict(name=humps.camelize(inner_stack), values=errors))
            else:
                traverse(errors, stack=inner_stack)

    traverse(parent["errors"])
    return fields


class CustomFieldErrorResolver(FallbackResolversSetter):
    def add_resolver_to_field(self, name: str, field_object: GraphQLField) -> None:
        if field_object.resolve is None and name == "errors":
            if (
                isinstance(field_object.type, GraphQLList)
                and field_object.type.of_type.name == "FieldError"
            ):
                field_object.resolve = resolve_field_errors


field_error_resolver = CustomFieldErrorResolver()



from .abc import Resolver
from .mixins import ListModelMixin, InputMixin, CreateModelMixin, UpdateModelMixin, DestroyModelMixin
from .model import GenericModelResolver, ModelResolver, ListModelResolver



from django.contrib.admin.options import get_content_type_for_model
from django.db import models

from . import exceptions
from ..filters import FilterMixin
from .mixins import (
    CreateModelMixin,
    DestroyModelMixin,
    ListModelMixin,
    UpdateModelMixin,
)
from .abc import Resolver


class GenericModelResolver(Resolver):
    """
    Model based resolver based off of DRF model views.
    """

    queryset = None
    lookup_arg = None
    lookup_field = "pk"

    def get_queryset(self):
        assert self.queryset is not None, (
            "'%s' should either include a `queryset` attribute, "
            "or override the `get_queryset()` method." % self.__class__.__name__
        )

        queryset = self.queryset
        if isinstance(queryset, models.query.QuerySet):
            # Ensure queryset is re-evaluated on each request.
            queryset = queryset.all()
        if self.config.get("nested", False):
            queryset = self.filter_nested_queryset(queryset)
        return queryset

    def filter_nested_queryset(self, queryset):
        # see if we can just use the info context and the name of the resolver field
        # to reverse the relation, fallback on parent_name
        try:
            return getattr(self.parent, self.info.field_name)
        except AttributeError:
            params = {}
            if self.config.get("parent_name", None) == "object_id":
                params["object_id"] = self.parent.id
                params["content_type"] = get_content_type_for_model(self.parent._meta.model)
            elif self.config.get("parent_name", None):
                params[self.config.get("parent_name")] = self.parent.id

            return queryset.filter(**params)

    def get_object(self):
        """
        Returns the object the resolver is displaying.

        You may want to override this if you need to provide non-standard
        queryset lookups.  Eg if objects are referenced using multiple
        keyword arguments in the url conf.
        """
        queryset = self.get_queryset()

        # Perform the lookup filtering.
        lookup_arg = self.lookup_arg or self.lookup_field

        assert lookup_arg in self.operation_kwargs, (
            "Expected resolver %s to be called with an argument "
            'named "%s". Fix your query arguments, or set the `.lookup_field` '
            "attribute on the resolver correctly." % (self.__class__.__name__, lookup_arg)
        )

        filter_kwargs = {self.lookup_field: self.operation_kwargs[lookup_arg]}
        obj = queryset.get(**filter_kwargs)
        # return null?
        # obj = get_object_or_404(queryset, **filter_kwargs)

        # May raise a permission denied
        if obj:
            self.check_object_permissions(self.info, obj)

        return obj

    # TODO: move out into its own mixin?
    def retrieve(self, parent, **kwargs):
        return self.get_object()


class ModelResolver(CreateModelMixin, UpdateModelMixin, DestroyModelMixin, FilterMixin, GenericModelResolver):
    pass


class ListModelResolver(ListModelMixin, GenericModelResolver):
    pass



"""
Mixins that are in conjunction with DRF serializers and resolver data
TODO: these are highly coupled to serializers, should they still be?
"""
import enum

from django.db.models.deletion import IntegrityError, ProtectedError



class ListModelMixin:
    """
    List a queryset.
    """

    pagination_class = None

    def list(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())

        page = self.paginate_queryset(queryset)
        if page is not None:
            return page

        return queryset

    @property
    def paginator(self):
        """
        The paginator instance associated with the view, or `None`.
        """
        if not hasattr(self, "_paginator"):
            if self.pagination_class is None:
                self._paginator = None
            else:
                self._paginator = self.pagination_class()
        return self._paginator

    def paginate_queryset(self, queryset, **kwargs):
        """
        Return a single page of results, or `None` if pagination is disabled.
        """
        if self.paginator is None:
            return None
        return self.paginator.paginate_queryset(
            queryset, self.request, view=self, **kwargs
        )


class InputMixin:
    input_arg = "input"
    convert_enums = True

    def get_input_arg(self):
        return self.input_arg

    def get_input_data(self):
        input_data = self.operation_kwargs.get(self.input_arg, {}).copy()
        # Get and convert any translated enums into their values
        # May be able to be moved to a serializer save point for models?
        # As having access to the enum may be useful in other contexts
        if self.convert_enums:
            for key, value in input_data.items():
                if isinstance(value, enum.Enum):
                    # Get and set actual value of enum
                    input_data[key] = value.value
        return input_data


class CreateModelMixin(InputMixin):
    def create(self, parent, **kwargs):
        serializer = self.get_serializer(data=self.get_input_data())
        valid = serializer.is_valid(raise_exception=False)
        if valid:
            obj = self.perform_create(serializer)
        else:
            obj = serializer.data
        return dict(success=valid, object=obj, errors=serializer.errors)

    def perform_create(self, serializer):
        return serializer.save()


class UpdateModelMixin:
    """
    Update a model instance.
    """

    def update(self, request, *args, **kwargs):
        partial = kwargs.pop("partial", False)
        instance = self.get_object()
        serializer = self.get_serializer(
            instance, data=self.get_input_data(), partial=partial
        )
        valid = serializer.is_valid(raise_exception=False)

        obj = instance
        if valid:
            obj = self.perform_update(serializer)

        if getattr(instance, "_prefetched_objects_cache", None):
            # If 'prefetch_related' has been applied to a queryset, we need to
            # forcibly invalidate the prefetch cache on the instance.
            instance._prefetched_objects_cache = {}

        return dict(success=valid, object=obj, errors=serializer.errors)

    def perform_update(self, serializer):
        return serializer.save()

    def partial_update(self, request, *args, **kwargs):
        kwargs["partial"] = True
        return self.update(request, *args, **kwargs)


class DestroyModelMixin:
    """
    Destroy a model instance.
    """

    def destroy(self, request, *args, **kwargs):
        instance = self.get_object()
        instance_id = instance.id
        count = False
        errors = {}
        try:
            count, _ = self.perform_destroy(instance)
        except (IntegrityError, ProtectedError) as e:
            errors = {"destroy": [str(e)]}
        return {"success": bool(count), "errors": errors, "object": {"id": instance_id}}

    def perform_destroy(self, instance):
        return instance.delete()
